/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer} data structure. This renderer mutates
   the {@link FrameBuffer}'s current viewport so that it is filled
   in with the rendered image of the scene represented by the
   {@link Scene} object.
<p>
   This implements our first rendering pipeline. It has just two
   pipeline stages.
   <ol>
   <li>The {@link Projection} transformation projects each {@link Vertex}
   from a {@link Model} into the {@link Camera}'s image plane.
   <li>The {@link Rasterize}r converts {@link LineSegment}s in the
   Camera's image plane into pixels in the {@link FrameBuffer}'s viewport.
   </ol>
   Notice that the first stage acts on the {@link Model}'s vertices, and
   the second stage acts on the {@link Model}'s line segments. That is,
   we project vertices and we rasterize line segments.
*/
public class Pipeline
{
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         if ( model.visible )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // 2. Rasterize each line segment into pixels.
            for (LineSegment ls : model2.lineSegmentList)
            {
               Rasterize.rasterize(model2, ls, fb);
            }
         }
      }
   }



   /**
      Render the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         System.out.println("=== Render Model ===================================");

         if ( model.visible )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // log interesting information to standard output
            logVertexList("0. Input    ", model2.vertexList);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // log interesting information to standard output
            logVertexList("1. Projected", model2.vertexList);

            for (LineSegment ls : model2.lineSegmentList)
            {
               // log interesting information to standard output
               logLineSegment("2. Rasterize", ls);

               // 2. Rasterize the line segment into pixels.
               Rasterize.rasterize(model2, ls, fb);
            }
            System.out.println();
         }
         else
         {
            System.out.println("   Hidden model.");
         }
         System.out.println("=== End Model ===================================");
      }
      System.out.println("=== End Scene ===================================");
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
   }
}
